"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.UtxoTransaction = void 0;
const genericMccTypes_1 = require("../../types/genericMccTypes");
const utxoTypes_1 = require("../../types/utxoTypes");
const errors_1 = require("../../utils/errors");
const utils_1 = require("../../utils/utils");
const utxoUtils_1 = require("../../utils/utxoUtils");
const TransactionBase_1 = require("../TransactionBase");
// Transaction types and their description
// - coinbase        : transaction that mints new coins
// - payment         : what you get from node
// - partial_payment : transaction with some vout of vins added to additional data
// - full_payment    : transaction with vouts for all vins added to additional data
//@Managed()
class UtxoTransaction extends TransactionBase_1.TransactionBase {
    get data() {
        return this.privateData;
    }
    get txid() {
        return this.data.txid;
    }
    get stdTxid() {
        return (0, utils_1.unPrefix0x)(this.txid);
    }
    get reference() {
        const references = [];
        for (const vo of this.data.vout) {
            if (vo.scriptPubKey.hex.substring(0, 2) == (0, utils_1.unPrefix0x)((0, utils_1.toHex)(utxoUtils_1.WordToOpcode.OP_RETURN))) {
                const dataSplit = vo.scriptPubKey.asm.split(" ");
                references.push(dataSplit[1]);
            }
        }
        return references;
    }
    get stdPaymentReference() {
        let paymentReference = this.reference.length === 1 ? (0, utils_1.prefix0x)(this.reference[0]) : "";
        if (!(0, utils_1.isValidBytes32Hex)(paymentReference)) {
            paymentReference = utils_1.ZERO_BYTES_32;
        }
        return paymentReference;
    }
    get unixTimestamp() {
        return this.data.mediantime || -1;
    }
    get sourceAddresses() {
        if ((0, utxoTypes_1.isCoinbase)(this.data)) {
            // Coinbase transactions mint coins
            return [undefined];
        }
        else if ((0, utxoTypes_1.hasPrevouts)(this.data)) {
            return this.data.vin.map((vin) => {
                return vin.prevout.scriptPubKey.address; // are we sure that every prevout has an address
            });
            //this means our assumptions on the transaction data are faulty
        }
        else
            throw (0, utils_1.MccError)(`transaction ${this.txid} that does not have prevout and is not coinbase`);
    }
    get receivingAddresses() {
        /* istanbul ignore if */
        // Transaction should always have vout array
        if (!this.data.vout) {
            return [];
        }
        return this.data.vout.map((vout) => vout.scriptPubKey.address);
    }
    /**
     * Returns value in minimal devisable units
     * @param value is stored in decimal with 1 representing the basic unit (BTC)
     * @returns
     */
    toBigIntValue(value) {
        if (value === undefined) {
            return BigInt(0);
        }
        else if (typeof value == "number") {
            let valueStr = value.toFixed(this.elementaryUnitsExponent);
            valueStr = valueStr.replace(".", "");
            return BigInt(valueStr);
        }
        else {
            const split = value.split(".");
            if (split.length == 1)
                return BigInt(value);
            else if (split.length == 2) {
                const valueStr = split[0].concat(split[1].padEnd(this.elementaryUnitsExponent, "0"));
                return BigInt(valueStr);
            }
            throw Error(`${value} is not a valid numeric string or number`);
        }
    }
    get fee() {
        switch (this.type) {
            case "payment": {
                let fee = BigInt(0);
                this.spentAmounts.forEach((addressAmount) => {
                    fee = fee + addressAmount.amount;
                });
                this.receivedAmounts.forEach((addressAmount) => {
                    fee = fee - addressAmount.amount;
                });
                return fee;
            }
            case "coinbase": {
                // Coinbase transactions mint coins
                return BigInt(0);
            }
        }
    }
    get feeSignerTotalAmount() {
        throw new Error("Method not implemented.");
    }
    get spentAmounts() {
        if ((0, utxoTypes_1.isCoinbase)(this.data)) {
            // Coinbase transactions mint coins
            return [
                {
                    amount: BigInt(0),
                },
            ];
        }
        else if ((0, utxoTypes_1.hasPrevouts)(this.data)) {
            return this.data.vin.map((mapper) => {
                let amount;
                if (mapper == undefined) {
                    amount = BigInt(0);
                }
                else {
                    amount = this.toBigIntValue(mapper.prevout.value || 0);
                }
                return {
                    address: mapper?.prevout?.scriptPubKey?.address,
                    amount: amount,
                    utxo: mapper?.vout,
                };
            });
        }
        else
            throw (0, utils_1.MccError)(`transaction ${this.txid} that does not have prevout and is not coinbase`);
    }
    get intendedSpentAmounts() {
        return this.spentAmounts;
    }
    get receivedAmounts() {
        return this.data.vout.map((vout) => {
            return {
                address: vout.scriptPubKey.address,
                amount: this.toBigIntValue(vout.value),
                utxo: vout.n,
            };
        });
    }
    get intendedReceivedAmounts() {
        return this.receivedAmounts;
    }
    get type() {
        if ((0, utxoTypes_1.isCoinbase)(this.data)) {
            return "coinbase";
        }
        else if ((0, utxoTypes_1.hasPrevouts)(this.data)) {
            return "payment";
        }
        else
            throw (0, utils_1.MccError)(`transaction ${this.txid} that does not have prevout and is not coinbase`);
    }
    get isNativePayment() {
        // On these chains there are no other types of transactions
        return true;
    }
    get successStatus() {
        return genericMccTypes_1.TransactionSuccessStatus.SUCCESS;
    }
    paymentSummary({ inUtxo, outUtxo }) {
        try {
            this.assertValidVinIndex(inUtxo);
        }
        catch (e) {
            return { status: TransactionBase_1.PaymentSummaryStatus.InvalidInUtxo };
        }
        try {
            this.assertValidVoutIndex(outUtxo);
        }
        catch (e) {
            return { status: TransactionBase_1.PaymentSummaryStatus.InvalidOutUtxo };
        }
        if (this.type === "coinbase") {
            return { status: TransactionBase_1.PaymentSummaryStatus.Coinbase };
        }
        const spentAmount = this.spentAmounts[inUtxo];
        if (!spentAmount.address) {
            return { status: TransactionBase_1.PaymentSummaryStatus.NoSpentAmountAddress };
        }
        const receiveAmount = this.receivedAmounts[outUtxo];
        if (!receiveAmount.address) {
            return { status: TransactionBase_1.PaymentSummaryStatus.NoReceiveAmountAddress };
        }
        // Extract addresses from input and output fields
        const sourceAddress = spentAmount.address;
        const receivingAddress = receiveAmount.address;
        // We will update this once we iterate over inputs and outputs if we have full transaction
        let oneToOne = true;
        let inFunds = BigInt(0);
        let returnFunds = BigInt(0);
        let outFunds = BigInt(0);
        let inFundsOfReceivingAddress = BigInt(0);
        for (const vinAmount of this.spentAmounts) {
            if (sourceAddress && vinAmount.address === sourceAddress) {
                inFunds = inFunds + vinAmount.amount;
            }
            if (receivingAddress && vinAmount.address === receivingAddress) {
                inFundsOfReceivingAddress = inFundsOfReceivingAddress + vinAmount.amount;
            }
            if (oneToOne && vinAmount.address != sourceAddress && vinAmount.address != receivingAddress) {
                oneToOne = false;
            }
        }
        for (const voutAmount of this.receivedAmounts) {
            if (sourceAddress && voutAmount.address === sourceAddress) {
                returnFunds = returnFunds + voutAmount.amount;
            }
            if (receivingAddress && voutAmount.address === receivingAddress) {
                outFunds = outFunds + voutAmount.amount;
            }
            // Outputs without address do not break one-to-one condition
            if (oneToOne && !voutAmount.address && voutAmount.amount >= BigInt(0)) {
                oneToOne = false;
            }
            if (oneToOne && voutAmount.address && voutAmount.address != sourceAddress && voutAmount.address != receivingAddress) {
                oneToOne = false;
            }
        }
        return {
            status: TransactionBase_1.PaymentSummaryStatus.Success,
            response: {
                blockTimestamp: this.unixTimestamp,
                transactionHash: this.stdTxid,
                sourceAddressHash: (0, utils_1.standardAddressHash)(sourceAddress),
                sourceAddressesRoot: this.sourceAddressesRoot,
                sourceAddress,
                receivingAddress,
                receivingAddressHash: (0, utils_1.standardAddressHash)(receivingAddress),
                spentAmount: inFunds - returnFunds,
                receivedAmount: outFunds - inFundsOfReceivingAddress,
                transactionStatus: this.successStatus,
                paymentReference: this.stdPaymentReference,
                intendedSourceAmount: inFunds - returnFunds,
                intendedReceivingAddressHash: (0, utils_1.standardAddressHash)(receivingAddress),
                intendedReceivingAddress: receivingAddress,
                intendedReceivingAmount: outFunds - inFundsOfReceivingAddress,
                oneToOne,
            },
        };
    }
    balanceDecreasingSummary(sourceAddressIndicator) {
        // We expect sourceAddressIndicator to be utxo vin index (as hex string)
        if (!(0, utils_1.isValidBytes32Hex)(sourceAddressIndicator)) {
            return { status: TransactionBase_1.BalanceDecreasingSummaryStatus.NotValidSourceAddressFormat };
        }
        const vinIndex = parseInt(sourceAddressIndicator, 16);
        if (isNaN(vinIndex)) {
            return { status: TransactionBase_1.BalanceDecreasingSummaryStatus.NotValidSourceAddressFormat };
        }
        try {
            this.assertValidVinIndex(vinIndex);
        }
        catch (e) {
            return { status: TransactionBase_1.BalanceDecreasingSummaryStatus.InvalidInUtxo };
        }
        const transactionType = this.type;
        switch (transactionType) {
            case "coinbase": {
                return { status: TransactionBase_1.BalanceDecreasingSummaryStatus.Coinbase };
            }
            case "payment": {
                const spentAmounts = this.spentAmounts;
                const spentAmount = spentAmounts[vinIndex];
                const sourceAddress = spentAmount.address;
                if (sourceAddress) {
                    let inFunds = BigInt(0);
                    let returnFunds = BigInt(0);
                    for (const vinAmount of this.spentAmounts) {
                        if (vinAmount.address === sourceAddress) {
                            inFunds = inFunds + vinAmount.amount;
                        }
                    }
                    for (const voutAmount of this.receivedAmounts) {
                        if (voutAmount.address === sourceAddress) {
                            returnFunds = returnFunds + voutAmount.amount;
                        }
                    }
                    return {
                        status: TransactionBase_1.BalanceDecreasingSummaryStatus.Success,
                        response: {
                            blockTimestamp: this.unixTimestamp,
                            transactionHash: this.stdTxid,
                            sourceAddressIndicator: sourceAddressIndicator,
                            sourceAddressHash: (0, utils_1.standardAddressHash)(sourceAddress),
                            sourceAddress: sourceAddress,
                            spentAmount: inFunds - returnFunds,
                            transactionStatus: this.successStatus,
                            paymentReference: this.stdPaymentReference,
                            isFull: true,
                        },
                    };
                }
                else {
                    // The input has no address, the type is wrongly extracted (Should not happen)
                    return { status: TransactionBase_1.BalanceDecreasingSummaryStatus.InvalidTransactionDataObject };
                }
                return { status: TransactionBase_1.BalanceDecreasingSummaryStatus.Success }; // TODO: implement
            }
            default:
                // exhaustive switch guard: if a compile time error appears here, you have forgotten one of the cases
                // eslint-disable-next-line @typescript-eslint/no-empty-function, @typescript-eslint/no-unused-vars
                ((_) => { })(transactionType);
        }
        // We didn't find the address we are looking for
        return { status: TransactionBase_1.BalanceDecreasingSummaryStatus.NoSourceAddress };
    }
    paymentNonexistenceSummary(outUtxo) {
        try {
            this.assertValidVoutIndex(outUtxo);
        }
        catch (e) {
            return { status: TransactionBase_1.PaymentNonexistenceSummaryStatus.InvalidOutUtxo };
        }
        if (this.type === "coinbase") {
            return { status: TransactionBase_1.PaymentNonexistenceSummaryStatus.Coinbase };
        }
        const receiveAmount = this.receivedAmounts[outUtxo];
        if (!receiveAmount.address) {
            return { status: TransactionBase_1.PaymentNonexistenceSummaryStatus.NoReceiveAmountAddress };
        }
        // Extract addresses from input and output fields
        const receivingAddress = receiveAmount.address;
        let outFunds = BigInt(0);
        let inFundsOfReceivingAddress = BigInt(0);
        for (const vinAmount of this.spentAmounts) {
            if (vinAmount.address === receivingAddress) {
                inFundsOfReceivingAddress = inFundsOfReceivingAddress + vinAmount.amount;
            }
        }
        for (const voutAmount of this.receivedAmounts) {
            if (voutAmount.address === receivingAddress) {
                outFunds = outFunds + voutAmount.amount;
            }
        }
        return {
            status: TransactionBase_1.PaymentNonexistenceSummaryStatus.Success,
            response: {
                blockTimestamp: this.unixTimestamp,
                transactionHash: this.stdTxid,
                receivingAddress,
                receivingAddressHash: (0, utils_1.standardAddressHash)(receivingAddress),
                sourceAddressesRoot: this.sourceAddressesRoot,
                receivedAmount: outFunds - inFundsOfReceivingAddress,
                transactionStatus: this.successStatus,
                paymentReference: this.stdPaymentReference,
                intendedReceivingAddressHash: (0, utils_1.standardAddressHash)(receivingAddress),
                intendedReceivingAddress: receivingAddress,
                intendedReceivingAmount: outFunds - inFundsOfReceivingAddress,
            },
        };
    }
    ///////////////////////////////
    //// Utxo specific methods ////
    ///////////////////////////////
    /**
     * Asserts whether the vin index is in valid range. If not, exception is thrown.
     * @param vinIndex vin index
     */
    assertValidVinIndex(vinIndex) {
        if (vinIndex < 0 || vinIndex >= this.sourceAddresses.length) {
            throw new errors_1.mccError(errors_1.mccErrorCode.InvalidParameter, Error("Invalid vin index"));
        }
    }
    /**
     * Asserts whether the vout index is in valid range. If not, exception is thrown.
     * @param voutIndex vout index
     */
    assertValidVoutIndex(voutIndex) {
        if (voutIndex < 0 || voutIndex >= this.receivingAddresses.length) {
            throw new errors_1.mccError(errors_1.mccErrorCode.InvalidParameter, Error("Invalid vout index"));
        }
    }
}
exports.UtxoTransaction = UtxoTransaction;
//# sourceMappingURL=UtxoTransaction.js.map